(ns metabase.api.field-test
  (:require [expectations :refer :all]
            [metabase.db :as db]
            (metabase.models [database :refer [Database]]
                             [field :refer [Field]]
                             [field-values :refer [FieldValues]]
                             [table :refer [Table]])
            [metabase.test.data :refer :all]
            [metabase.test.data.users :refer :all]
            [metabase.test.util :as tu]))

;; Helper Fns

(defn- db-details []
  (tu/match-$ (db)
    {:created_at         $
     :engine             "h2"
     :caveats            nil
     :points_of_interest nil
     :id                 $
     :updated_at         $
     :name               "test-data"
     :is_sample          false
     :is_full_sync       true
     :organization_id    nil
     :description        nil
     :features           (mapv name (metabase.driver/features (metabase.driver/engine->driver :h2)))}))


;; ## GET /api/field/:id
(expect
  (tu/match-$ (Field (id :users :name))
    {:description        nil
     :table_id           (id :users)
     :raw_column_id      $
     :table              (tu/match-$ (Table (id :users))
                           {:description             nil
                            :entity_type             nil
                            :visibility_type         nil
                            :db                      (db-details)
                            :schema                  "PUBLIC"
                            :name                    "USERS"
                            :display_name            "Users"
                            :rows                    15
                            :updated_at              $
                            :entity_name             nil
                            :active                  true
                            :id                      (id :users)
                            :db_id                   (id)
                            :caveats                 nil
                            :points_of_interest      nil
                            :show_in_getting_started false
                            :raw_table_id            $
                            :created_at              $})
     :special_type       "name"
     :name               "NAME"
     :display_name       "Name"
     :caveats            nil
     :points_of_interest nil
     :updated_at         $
     :last_analyzed      $
     :active             true
     :id                 (id :users :name)
     :field_type         "info"
     :visibility_type    "normal"
     :position           0
     :preview_display    true
     :created_at         $
     :base_type          "TextField"
     :fk_target_field_id nil
     :parent_id          nil})
  ((user->client :rasta) :get 200 (format "field/%d" (id :users :name))))


;; ## GET /api/field/:id/summary
(expect [["count" 75]      ; why doesn't this come back as a dictionary ?
         ["distincts" 75]]
  ((user->client :rasta) :get 200 (format "field/%d/summary" (id :categories :name))))


;; ## PUT /api/field/:id

(defn- simple-field-details [field]
  (select-keys field [:name :display_name :description :visibility_type :special_type]))

;; test that we can do basic field update work, including unsetting some fields such as special-type
(expect
  [{:name            "Field Test"
    :display_name    "Field Test"
    :description     nil
    :special_type    nil
    :visibility_type :normal}
   {:name            "Field Test"
    :display_name    "yay"
    :description     "foobar"
    :special_type    :name
    :visibility_type :sensitive}
   {:name            "Field Test"
    :display_name    "yay"
    :description     nil
    :special_type    nil
    :visibility_type :sensitive}]
  (tu/with-temp* [Database [{database-id :id} {:name               "Field Test"
                                               :engine             :yeehaw
                                               :caveats            nil
                                               :points_of_interest nil
                                               :details            {}
                                               :is_sample          false}]
                  Table    [{table-id :id}    {:name   "Field Test"
                                               :db_id  database-id
                                               :active true}]
                  Field    [{field-id :id}    {:table_id        table-id
                                               :name            "Field Test"
                                               :base_type       :TextField
                                               :field_type      :info
                                               :special_type    nil
                                               :active          true
                                               :preview_display true
                                               :position        1}]]
    (let [original-val (simple-field-details (Field field-id))]
      ;; set it
      ((user->client :crowberto) :put 200 (format "field/%d" field-id) {:name            "something else"
                                                                        :display_name    "yay"
                                                                        :description     "foobar"
                                                                        :special_type    :name
                                                                        :visibility_type :sensitive})
      (let [updated-val (simple-field-details (Field field-id))]
        ;; unset it
        ((user->client :crowberto) :put 200 (format "field/%d" field-id) {:description  nil
                                                                          :special_type nil})
        [original-val
         updated-val
         (simple-field-details (Field field-id))]))))

;; when we set the special-type from :fk to something else, make sure fk_target_field_id is set to nil
(expect
  [true
   nil]
  (tu/with-temp* [Database [{database-id :id} {:name               "Field Test"
                                               :engine             :yeehaw
                                               :caveats            nil
                                               :points_of_interest nil
                                               :details            {}
                                               :is_sample          false}]
                  Table    [{table-id :id}    {:name   "Field Test"
                                               :db_id  database-id
                                               :active true}]
                  Field    [{field-id1 :id}   {:table_id        table-id
                                               :name            "Target Field"
                                               :base_type       :TextField
                                               :field_type      :info
                                               :special_type    :id
                                               :active          true
                                               :preview_display true
                                               :position        1}]
                  Field    [{field-id :id}    {:table_id           table-id
                                               :name               "Field Test"
                                               :base_type          :TextField
                                               :field_type         :info
                                               :special_type       :fk
                                               :fk_target_field_id field-id1
                                               :active             true
                                               :preview_display    true
                                               :position           1}]]
    (let [original-val (boolean (db/select-one-field :fk_target_field_id Field, :id field-id))]
      ;; unset the :fk special-type
      ((user->client :crowberto) :put 200 (format "field/%d" field-id) {:special_type :name})
      [original-val
       (db/select-one-field :fk_target_field_id Field, :id field-id)])))

;; check that you can't set a field to :timestamp_seconds if it's not of a proper base_type
(expect
  ["Invalid Request."
   nil]
  (tu/with-temp* [Database [{database-id :id}]
                  Table    [{table-id :id} {:db_id database-id}]
                  Field    [{field-id :id} {:table_id table-id}]]
    [((user->client :crowberto) :put 400 (str "field/" field-id) {:special_type :timestamp_seconds})
     (db/select-one-field :special_type Field, :id field-id)]))


(defn- field->field-values
  "Fetch the `FieldValues` object that corresponds to a given `Field`."
  [table-kw field-kw]
  (FieldValues :field_id (id table-kw field-kw)))

(defn- field-values-id [table-key field-key]
  (:id (field->field-values table-key field-key)))

;; ## GET /api/field/:id/values
;; Should return something useful for a field that has special_type :category
(expect
  {:field_id              (id :venues :price)
   :human_readable_values {}
   :values                [1 2 3 4]
   :id                    (field-values-id :venues :price)}
  (do
    ;; clear out existing human_readable_values in case they're set
    (db/update! FieldValues (field-values-id :venues :price)
      :human_readable_values nil)
    ;; now update the values via the API
    (-> ((user->client :rasta) :get 200 (format "field/%d/values" (id :venues :price)))
        (dissoc :created_at :updated_at))))

;; Should return nothing for a field whose special_type is *not* :category
(expect
  {:values                {}
   :human_readable_values {}}
  ((user->client :rasta) :get 200 (format "field/%d/values" (id :venues :id))))


;; ## POST /api/field/:id/value_map_update

;; Check that we can set values
(expect
  [{:status "success"}
   {:field_id              (id :venues :price)
    :human_readable_values {:1 "$"
                            :2 "$$"
                            :3 "$$$"
                            :4 "$$$$"}
    :values                [1 2 3 4]
    :id                    (field-values-id :venues :price)}]
  [((user->client :crowberto) :post 200 (format "field/%d/value_map_update" (id :venues :price)) {:values_map {:1 "$"
                                                                                                               :2 "$$"
                                                                                                               :3 "$$$"
                                                                                                               :4 "$$$$"}})
   (-> ((user->client :rasta) :get 200 (format "field/%d/values" (id :venues :price)))
       (dissoc :created_at :updated_at))])

;; Check that we can unset values
(expect
  [{:status "success"}
   (tu/match-$ (FieldValues :field_id (id :venues :price))
     {:field_id              (id :venues :price)
      :human_readable_values {}
      :values                [1 2 3 4]
      :id                    (field-values-id :venues :price)})]
  [(do (db/update! FieldValues (:id (field->field-values :venues :price))
         :human_readable_values {:1 "$" ; make sure they're set
                                 :2 "$$"
                                 :3 "$$$"
                                 :4 "$$$$"})
       ((user->client :crowberto) :post 200 (format "field/%d/value_map_update" (id :venues :price)) {:values_map {}}))
   (-> ((user->client :rasta) :get 200 (format "field/%d/values" (id :venues :price)))
       (dissoc :created_at :updated_at))])

;; Check that we get an error if we call value_map_update on something that isn't a category
(expect "You can only update the mapped values of a Field whose 'special_type' is 'category'/'city'/'state'/'country' or whose 'base_type' is 'BooleanField'."
  ((user->client :crowberto) :post 400 (format "field/%d/value_map_update" (id :venues :id))
   {:values_map {:1 "$"
                 :2 "$$"
                 :3 "$$$"
                 :4 "$$$$"}}))
